""" Handles the GPT API and the conversation state. """
import json
import re
import time
from pathlib import Path
from PIL.Image import Image

from justai.tools.cache import cached_llm_response
from justai.model.message import Message
from justai.models.modelfactory import ModelFactory


class Model:
    def __init__(self, model_name: str, **kwargs):
        
        # Model parameters
        self.model = ModelFactory.create(model_name, **kwargs)

        # Parameters to save the current conversation
        self.save_dir = Path(__file__).resolve().parent / 'saves'
        self.message_memory = 20  # Number of messages to remember. Limits token usage.
        self.messages = []  # List of Message objects
        self.tools = []  # List of tools to use / functions to call
        self.functions = {}  # The actual functions to call with key the name of the function and as value the function

        self.input_token_count = 0
        self.output_token_count = 0
        self.last_response_time = 0
        
        self.logger = None
        
    def __setattr__(self, name, value):
        if name not in self.__dict__ and hasattr(self, 'model') and name in self.model.model_params:
            # Not an existing property model but a model_params property. Set it in model_params
            self.model.model_params[name] = value
        else:
            # Update the property as intended
            super().__setattr__(name, value)

    @classmethod
    def from_json(cls, model_name, model_data, **kwargs):
        """ Creates an model from a json string. Usefull in stateless environments like a web page """
        model = cls(model_name, **kwargs)
        dictionary = json.loads(model_data)
        for key, value in dictionary.items():
            match key:
                case 'messages':
                    model.messages = [Message.from_dict(m) for m in value]
                case _:
                    model.__setattr__(key, value)
        return model

    def set_api_key(self, key: str):
        """ Used when using Aigent from a browser where the user has to specify a key """
        self.model.set('api_key', key)

    @property
    def system(self):  # This function can be overwritten by child classes to make the system message dynamic
        return self.model.system_message

    @system.setter
    def system(self, value):
        self.model.system_message = value

    @property
    def cached_prompt(self): 
        if hasattr(self.model, 'cached_prompt'):
            return self.model.cached_prompt
        raise AttributeError("Model does not support cached_prompt")

    @cached_prompt.setter
    def cached_prompt(self, value):
        if hasattr(self.model, 'cached_prompt'):
            self.model.cached_prompt = value
        else:
            raise AttributeError("Model does not support cached_prompt")

    @property
    def cache_creation_input_tokens(self):
        if hasattr(self.model, 'cache_creation_input_tokens'):
            return self.model.cache_creation_input_tokens
        raise AttributeError("Model does not support cache_creation_input_tokens")
    
    @property
    def cache_read_input_tokens(self):
        if hasattr(self.model, 'cache_read_input_tokens'):
            return self.model.cache_read_input_tokens
        raise AttributeError("Model does not support cache_read_input_tokens")
        
    def reset(self):
        self.messages = []

    def append_messages(self, prompt: str,
                        images: [list[str] | list[bytes] | list[Image] | str | bytes | Image | None] = None):
        if images:
            if not isinstance(images, list):
                images = [images]
        else:
            images = []
        self.messages.append(Message('user', prompt, images))
        return self.messages

    def add_function(self, func: callable, description: str, parameters: dict, required_parameters=None):
        """ Adds a function to the model.
        name: name of the function
        description: description of the function
        parameters: dictionary with parameter description as key and parameter type as value """
        tool = {
            "type": "function",
            "function": {
                "name": func.__name__,
                "description": description,
                "parameters": {
                    "type": "object",
                    "properties": {
                        re.sub(r"\W+", "", description).lower(): {
                            "type": _type.__name__ if isinstance(_type, type) else str(_type),
                            "description": description,
                        }
                        for description, _type in parameters.items()
                    },
                    "required": required_parameters or [],
                    "additionalProperties": False
                }
            }
        }
        self.tools.append(tool)
        self.functions[func.__name__] = func

    def get_messages(self) -> list[Message]:
        return self.messages[-self.message_memory:]

    def last_token_count(self):
        return self.input_token_count, self.output_token_count, self.input_token_count + self.output_token_count

    def chat(self, prompt, *, images: [list[str] | list[bytes] | list[Image] | str | bytes | Image | None] = None,
             return_json=False, response_format=None, cached=True):
        start_time = time.time()
        if images and not isinstance(images, list):
            images = [images]
        self.append_messages(prompt, images)

        model_response = cached_llm_response(self.model, self.get_messages(), self.tools, return_json=return_json,
                                             response_format=response_format, use_cache=cached)
        result, self.input_token_count, self.output_token_count, tool_use = model_response
        calls = 0  # Safety to prevent infinite loop
        while tool_use and calls < 3:
            calls += 1
            function = self.functions.get(tool_use["function_to_call"])
            if not function:
                raise ValueError(f"Function {tool_use["function_to_call"]} not found")

            # Run the tool/function
            tool_use['function_result'] = function(*tool_use['function_parameters'].values())

            # Add a model specific message with the tool use result to the conversation
            self.messages.append(self.model.tool_use_message(tool_use))

            model_response = cached_llm_response(
                self.model,
                self.get_messages(),
                self.tools,
                return_json=return_json,
                response_format=response_format,
                use_cache=cached,
            )
            result, self.input_token_count, self.output_token_count, tool_use = model_response

        # Add the result to the messages
        self.messages.append(Message('assistant', str(result)))
        self.last_response_time = time.time() - start_time
        return result
    
    async def chat_async(self, prompt, *,
                         images: [list[str] | list[bytes] | list[Image] | str | bytes | Image | None] = None):
        if images and not isinstance(images, list):
            images = [images]
        self.append_messages(prompt, images)
        for word, _ in self.model.chat_async(messages=self.get_messages()):
            if word:
                yield word

    async def chat_async_reasoning(self, prompt, *,
                         images: [list[str] | list[bytes] | list[Image] | str | bytes | Image | None] = None):
        """ Same as chat_async but returns the reasoning content as well
        """
        if images and not isinstance(images, list):
            images = [images]
        self.append_messages(prompt, images)
        for word, reasoning_content in self.model.chat_async(messages=self.get_messages()):
            if word or reasoning_content:
                yield word, reasoning_content

    def after_response(self):
        # content is in messages[-1]['completion']['choices'][0]['message']['content']
        return  # Can be overridden

    def token_count(self, text: str):
        return self.model.token_count(text)



""" Implementation of the Anthropic models. 

Feature table:
    - Async chat:       YES (1)
    - Return JSON:      YES
    - Structured types: NO
    - Token count:      YES
    - Image support:    YES
    - Tool use:         not yet
    
Models:
Claude 3 Opus:	 claude-3-opus-20240229
Claude 3 Sonnet: claude-3-5-sonnet-20240620
Claude 3 Haiku:  claude-3-haiku-20240307

Supported parameters:
max_tokens: int (default 800)
temperature: float (default 0.8)

(1) In contrast to Model.chat, Model.chat_async cannot return json and does not return input and output token counts

"""

import json
import os
from typing import Any

from anthropic import Anthropic, AsyncAnthropic, APIConnectionError, AuthenticationError, PermissionDeniedError, \
    APITimeoutError, RateLimitError, BadRequestError, Timeout
from dotenv import dotenv_values
from google.api_core.exceptions import InternalServerError

from justai.model.message import Message
from justai.models.basemodel import BaseModel, identify_image_format_from_base64, ConnectionException, AuthorizationException, \
    ModelOverloadException, RatelimitException, BadRequestException, GeneralException
from justai.tools.display import ERROR_COLOR, color_print


class AnthropicModel(BaseModel):
    def __init__(self, model_name: str, params: dict):
        system_message = f"You are {model_name}, a large language model trained by Anthropic."
        super().__init__(model_name, params, system_message)
        self.cached_prompt = None

        # Authentication
        if "ANTHROPIC_API_KEY" in params:
            api_key = params["ANTHROPIC_API_KEY"]
            del params["ANTHROPIC_API_KEY"]
        else:
            api_key = os.getenv("ANTHROPIC_API_KEY") or dotenv_values()["ANTHROPIC_API_KEY"]
        if not api_key:
            color_print(
                "No Anthropic API key found. Create one at https://console.anthropic.com/settings/keys and "
                + "set it in the .env file like ANTHROPIC_API_KEY=here_comes_your_key.",
                color=ERROR_COLOR,
            )

        # Client
        factory = AsyncAnthropic if params.get("async") else Anthropic
        self.client = factory(api_key=api_key, timeout=Timeout(10.0))

        # Required model parameters
        if "max_tokens" not in params:
            params["max_tokens"] = 800

    def chat(self, messages: list[Message], tools, return_json: bool, response_format, log=None) \
            -> tuple[str | Any, int, int, dict[str, object | str] | dict[str, Any]]:
        if response_format:
            raise NotImplementedError("Anthropic does not support response_format")

        antr_messages = transform_messages(messages, return_json)
        antr_tools = transform_tools(tools)
        system_message = self.cached_system_message() if self.cached_prompt else self.system_message
        try:
            message = self.client.messages.create(model=self.model_name, system=system_message, messages=antr_messages,
                                                  tools=antr_tools, **self.model_params)
        except APIConnectionError as e:
            raise ConnectionException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            raise AuthorizationException(e)
        except InternalServerError as e:
            raise ModelOverloadException(e)
        except RateLimitError as e:
            raise RatelimitException(e)
        except BadRequestError as e:
            raise BadRequestException(e)
        except Exception as e:
            raise GeneralException(e)

        # Text content
        response_str = message.content[0].text
        if return_json:
            response_str = response_str.split("</json>")[0]  # !!
            try:
                response = json.loads(response_str, strict=False)
            except json.decoder.JSONDecodeError:
                print("ERROR DECODING JSON, RESPONSE:", response_str)
                response = response_str
        else:
            response = response_str

        # Tool use
        # if message.stop_reason == "tool_use":
        #     for c in message.content:
        #         if isinstance(c, ToolUseBlock):
        #             tool_use = {'function_to_call': c.name,
        #                         'function_parameters': c.input,
        #                         'call_id': c.id}
        #             break  # For now, only one function call is supported
        # else:
        #     tool_use = {}
        tool_use = {}

        # Token count
        input_tokens = message.usage.input_tokens
        output_tokens = message.usage.output_tokens
        if self.cached_prompt:
            self.cache_creation_input_tokens = message.usage.cache_creation_input_tokens
            self.cache_read_input_tokens = message.usage.cache_read_input_tokens
        else:
            self.cache_creation_input_tokens = self.cache_read_input_tokens = 0

        return response, input_tokens, output_tokens, tool_use

    def chat_async(self, messages: list[Message]) -> [str, str]:
        try:
            # stream = self.client.messages.create(
            #     model=self.model_name,
            #     max_tokens=self.model_params["max_tokens"],
            #     temperature=self.model_params.get('temperature', 0.8),
            #     system=self.system_message,
            #     messages=transform_messages(messages, return_json=False),
            #     stream=True,
            # )
            stream = self.client.messages.create(
                model=self.model_name,
                system=self.system_message,
                messages=transform_messages(messages, return_json=False),
                stream=True,
                **self.model_params
            )
        except APIConnectionError as e:
            raise ConnectionException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            raise AuthorizationException(e)
        except APITimeoutError as e:
            raise ModelOverloadException(e)
        except RateLimitError as e:
            raise RatelimitException(e)
        except BadRequestError as e:
            raise BadRequestException(e)
        except Exception as e:
            raise GeneralException(e)

        for event in stream:
            if hasattr(event, "delta") and hasattr(event.delta, "text"):
                yield event.delta.text, None   # 2nd parameter is reasoning_content. Not available yet for Anthropic

    def cached_system_message(self) -> list[dict]:
        return [
                  {
                    "type": "text",
                    "text": self.system_message,
                  },
                  {
                    "type": "text",
                    "text": self.cached_prompt,
                    "cache_control": {"type": "ephemeral"}
                  }
                ]

    @staticmethod  # !! Veranderen als deze niet anders wordt dan die van OpenAI
    def tool_use_message(tool_use) -> Message:
        return Message(role='user', content='', tool_use=tool_use)

    def token_count(self, text: str) -> int:
        messages = transform_messages([Message("user", text)], return_json=False)
        return self.client.beta.messages.count_tokens(model=self.model_name, messages=messages)


def transform_messages(messages: list[Message], return_json: bool) -> list[dict]:
    # Anthropic requires the first message to be a user message
    msgs = messages[next(i for i, message in enumerate(messages) if message.role == "user"):]

    if msgs and return_json:
        msgs += [Message("assistant", "<json>")]
    result = [create_anthropic_message(msg) for msg in msgs]
    return result


def transform_tools(tools: list[dict]) -> list[dict]:
    """
    At Anthropic tools work like this:
    tools=[
        {
            "name": "get_weather",
            "description": "Get the current weather in a given location",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    }
                },
                "required": ["location"],
            },
        }
    ],
    """
    def transform_tool(tool:dict) -> dict:
        if "function" in tool:
            tool["name"] = tool["function"]["name"]
            tool["description"] = tool["function"]["description"]
            tool["input_schema"] = tool["function"]["parameters"]
            del tool["function"]
        if "type" in tool:
            del tool["type"]
        if "function" in tool:
            del tool["function"]
        return tool

    return [transform_tool(tool) for tool in tools]


def create_anthropic_message(message: Message):

    content = []
    role = message.role

    for img in message.images:
        base64img = Message.to_base64_image(img)
        mime_type = identify_image_format_from_base64(base64img)
        content += [
            {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": mime_type,
                    "data": base64img,
                },
            }
        ]

    if message.content:
        content += [{"type": "text", "text": message.content}]

    if message.tool_use:
        _input = message.tool_use["function_result"]
        if not isinstance(input, dict): # Anthropic requires input to be a dict
            _input = {'data': _input}
        content += [{
            "type": "tool_use",
            "id": message.tool_use["call_id"],
            "name": message.tool_use["function_to_call"],
            "input": _input,
        }]
        role = 'assistant' # Anthropic requires the role to be assistant when using tools

    return {"role": role, "content": content}



import unittest
import tempfile
import os
from justai.tools.prompts import set_prompt_file, add_prompt_file, get_prompt

class TestPrompts(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.prompt_file = os.path.join(self.temp_dir, "test_prompts.toml")
        with open(self.prompt_file, "w") as f:
            f.write('test_prompt = "This is a test prompt {variable}"\n')

    def tearDown(self):
        os.remove(self.prompt_file)
        os.rmdir(self.temp_dir)

    def test_set_and_get_prompt(self):
        set_prompt_file(self.prompt_file)
        prompt = get_prompt("test_prompt", variable="value")
        self.assertEqual(prompt, "This is a test prompt value")

    def test_add_prompt_file(self):
        set_prompt_file(self.prompt_file)
        new_prompt_file = os.path.join(self.temp_dir, "new_prompts.toml")
        with open(new_prompt_file, "w") as f:
            f.write('new_prompt = "This is a new prompt {variable}"\n')
        add_prompt_file(new_prompt_file)
        prompt = get_prompt("new_prompt", variable="value")
        self.assertEqual(prompt, "This is a new prompt value")
        os.remove(new_prompt_file)

if __name__ == '__main__':
    unittest.main()


import base64
from abc import ABC, abstractmethod

from justai.model.message import Message


class ConnectionException(Exception):
    pass

class AuthorizationException(Exception):
    pass

class ModelOverloadException(Exception):
    pass

class RatelimitException(Exception):
    pass

class BadRequestException(Exception):
    pass

class TimeoutException(Exception):
    pass

class GeneralException(Exception):
    pass


class BaseModel(ABC):

    @abstractmethod
    def __init__(self, model_name: str, params: dict, system_message: str):
        """ Model implemention should create attributes for all supported parameters """
        self.model_name = model_name
        self.model_params = params  # Specific parameters for specific models like temperature
        self.system_message = system_message
        self.debug = params.get('debug', False)

    def set(self, key: str, value):
        if not hasattr(self, key):
            raise (AttributeError(f"Model has no attribute {key}"))
        setattr(self, key, value)

    @abstractmethod
    def chat(self, messages: list[Message], tools: list, return_json: bool, response_format) \
            -> tuple[[str | object], int, int]:
        pass

    @abstractmethod
    def chat_async(self, messages: list[Message]) -> str:
        pass

    @abstractmethod
    def token_count(self, text: str) -> int:
        pass


def identify_image_format_from_base64(encoded_data: str) -> str:
    raw_data = base64.b64decode(encoded_data)[:8]

    # Magic numbers and corresponding mime types for each image format
    formats = {
        b'\xff\xd8\xff': 'image/jpeg',  # JPEG files
        b'\x89PNG\r\n\x1a\n': 'image/png',  # PNG files
        b'GIF87a': 'image/gif',  # GIF files (version 87a)
        b'GIF89a': 'image/gif',  # GIF files (version 89a)
        b'BM': 'image/bmp',  # BMP files
        b'II*\x00': 'image/tiff',  # TIFF files (little endian)
        b'MM\x00*': 'image/tiff',  # TIFF files (big endian)
        b'RIFF': 'image/webp',  # WebP files (part of RIFF)
        b'<svg': 'image/svg+xml',  # SVG files (this is a rough check; SVG is a text-based format)
        b'<!DOCT': 'image/svg+xml',  # SVG files with DOCTYPE declaration
    }

    # Check the raw data against known magic numbers
    for magic, mime_type in formats.items():
        if raw_data.startswith(magic):
            return mime_type

    return 'unknown/unknown'



import unittest
from justai.tools.cache import CachDB, cached_llm_response
from unittest.mock import patch, MagicMock

class TestCache(unittest.TestCase):
    def setUp(self):
        self.cache = CachDB()

    def tearDown(self):
        self.cache.clear()
        self.cache.close()

    def test_write_and_read(self):
        self.cache.write("test_key", ("test_value", 10, 5))
        result = self.cache.read("test_key")
        self.assertEqual(result, ("test_value", 10, 5))

    @patch('justai.tools.cache.CachDB')
    def test_cached_llm_response(self, mock_cache):
        mock_model = MagicMock()
        mock_model.chat.return_value = ("Test response", 10, 5)
        mock_cache_instance = mock_cache.return_value
        mock_cache_instance.read.return_value = None

        result = cached_llm_response(mock_model, [], [], False)
        self.assertEqual(result, ("Test response", 10, 5))
        mock_cache_instance.write.assert_called_once()

if __name__ == '__main__':
    unittest.main()


import hashlib
import os
import json
import sqlite3
from pathlib import Path

from justdays import Day

from justai.model.message import Message


def cached_llm_response(model, messages: list[Message], tools: list, return_json: bool, response_format=None,
                        use_cache=True, max_retries=None) -> tuple[[str | object], int, int, dict]:
    if not use_cache:
        return model.chat(messages, tools, return_json, response_format, max_retries)

    hashcode = recursive_hash((model, messages, return_json))
    cachedb = CachDB()
    result = cachedb.read(hashcode)
    if result:
        if return_json:
            return json.loads(result[0]), result[1], result[2], result[3]
        return result
    result = model.chat(messages, tools, return_json, response_format, max_retries)
    if return_json:
        cachedb.write(hashcode, (json.dumps(result[0]), result[1], result[2], result[3]))
    else:
        cachedb.write(hashcode, result)
    return result


cache_dir = ''
cache_file = 'llmcache.db'


def set_cache_dir(_dir):
    global cache_dir
    cache_dir = _dir


class CachDB:
    _instance = None

    def __new__(cls, *args, **kwargs):  # Make this class a singleton
        if cls._instance is None:
            cls._instance = super(CachDB, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        global cache_dir, cache_file
        dir_ = os.getenv('CACHE_DIR', cache_dir) or Path(__file__).resolve().parent
        self.db_path = os.path.join(dir_, cache_file)
        self.conn = sqlite3.connect(self.db_path)

        self.cursor = self.conn.cursor()
        self.cursor.execute('''CREATE TABLE IF NOT EXISTS cache (
                                    hashkey VARCHAR(32) PRIMARY KEY,
                                    value TEXT,
                                    tokens_in INT,
                                    tokens_out INT,
                                    valid_until DATETIME)''')
        self.cursor.execute('DELETE FROM cache WHERE valid_until < ?', (str(Day()),))
        self.conn.commit()

    def write(self, key: str, llm_response: tuple[str, int, int], valid_until: str = ''):
        if not valid_until:
            valid_until = str(Day().plus_months(1))
        value, tokens_in, tokens_out = llm_response
        try:
            self.cursor.execute('''INSERT INTO cache (hashkey, value, tokens_in, tokens_out, valid_until) 
                                    VALUES (?, ?, ?, ?, ?)''', (key, value, tokens_in, tokens_out, valid_until))
            self.conn.commit()
        except sqlite3.ProgrammingError:
            pass  # Something went wrong. Whatever, just don't add to the cache but never crash

    def read(self, key):
        self.cursor.execute("SELECT * FROM cache WHERE hashkey = ?", (key,))
        result = self.cursor.fetchone()
        if result:
            return result[1], result[2], result[3]

    def clear(self):
        try:
            self.cursor.execute('DELETE FROM cache')
            self.conn.commit()
        except sqlite3.ProgrammingError:
            pass  # Something went wrong. Whatever, just don't delete the cache but never crash

    def close(self):
        self.conn.close()


def recursive_hash(value, depth=0, ignore_params=[]):
    """Hash primitives recursively with maximum depth. Via https://docs.sweep.dev/blogs/file-cache"""
    if depth > 6:
        return hash_code("max_depth_reached")
    if isinstance(value, (int, float, str, bool, bytes)):
        return hash_code(str(value))
    if isinstance(value, (list, tuple)):
        return hash_code("".join([recursive_hash(item, depth + 1, ignore_params) for item in value]))
    if isinstance(value, dict):
        return hash_code(
            "".join(
                [
                    recursive_hash(key, depth + 1, ignore_params)
                    + recursive_hash(val, depth + 1, ignore_params)
                    for key, val in value.items()
                    if key not in ignore_params
                ]
            ))
    if hasattr(value, "__dict__") and value.__class__.__name__ not in ignore_params:
        return recursive_hash(value.__dict__, depth + 1, ignore_params)
    return hash_code("unknown")


def hash_code(code):
    return hashlib.md5(code.encode()).hexdigest()



import os
import unittest
from unittest.mock import patch
from justai.model.model import Model


class TestAgent(unittest.TestCase):
    def setUp(self):
        # Lijst van netwerken om te testen
        self.networks = ["gpt-4o", "gpt-4o-mini", "claude-3-opus-20240229", "claude-3-5-sonnet-20240620", 
                         "gemini-1.5-pro", "gemini-1.5-pro"]
        self.models = [Model(network) for network in self.networks]

    def test_chat(self):
        for i, model in enumerate(self.models):
            with self.subTest(network=self.networks[i]):
                with patch('justai.tools.cache.cached_llm_response') as mock_response:
                    mock_response.return_value = ("Test response", 10, 5)
                    response = model.chat("Test prompt")
                    self.assertEqual(isinstance(response, str), True)
                    self.assertGreater(len(response), 0)

    def test_append_messages(self):
        for i, model in enumerate(self.models):
            with self.subTest(network=self.networks[i]):
                model.append_messages("Test message")
                self.assertEqual(len(model.messages), 1)
                self.assertEqual(model.messages[0].content, "Test message")

    def test_reset(self):
        for i, model in enumerate(self.models):
            with self.subTest(network=self.networks[i]):
                model.append_messages("Test message")
                model.reset()
                self.assertEqual(len(model.messages), 0)

if __name__ == '__main__':
    os.chdir(os.path.join(os.path.dirname(__file__), '..'))
    unittest.main()


""" Implementation of the Anthropic models. 
Uses the Google AI library. TODO: Add support for Google Vertex library one day

Feature table:
    - Async chat:       YES (1)
    - Return JSON:      YES
    - Structured types: YES, via Python type definition
    - Token counter:    YES
    - Image support:    YES 
    - Tool use:         NO (not yet)

Models:
gemini-1.5-flash -> gemini-1.5-flash-001
gemini-1.5-pro -> gemini-1.5-pro-001

Supported parameters: 
max_tokens: int (default: None)
temperature: float (default:None)
stop_sequences: list[str] (default: None)
candidate_count: int (default: None)
topP: float (default: None)
topK: int (default: None)

(1) In contrast to Model.chat, Model.chat_async cannot return json and does not return input and output token counts

"""
import json
import os
import absl.logging
from contextlib import contextmanager
import time

from dotenv import dotenv_values
import google
import google.generativeai as genai

from justai.model.message import Message
from justai.models.basemodel import BaseModel
from justai.tools.display import ERROR_COLOR, color_print


class GoogleModel(BaseModel):

    def __init__(self, model_name: str, params: dict):
        system_message = f"You are {model_name}, a large language model trained by Google."
        super().__init__(model_name, params, system_message)

        # Authentication
        api_key = params.get("GOOGLE_API_KEY") or os.getenv("GOOGLE_API_KEY") or dotenv_values()["GOOGLE_API_KEY"]
        if api_key:
            genai.configure(api_key=api_key)
        else:
            color_print("No Google API key found. Create one at https://aistudio.google.com/app/apikey and " +
                        "set it in the .env file like GOOGLE_API_KEY=here_comes_your_key.", color=ERROR_COLOR)

        # Client
        self.client = genai.GenerativeModel(model_name, system_instruction=system_message)

    def chat(self, messages: list[Message], tools, return_json: bool, response_format, max_retries=None, log=None) \
            -> tuple[[str | object], int, int]:
        if tools:
            raise NotImplementedError("tools are not supported (yet)")

        google_messages = transform_messages(messages, return_json)
        chat = self.client.start_chat(history=google_messages[:-1])

        with temporary_verbosity(absl.logging.WARNING):
            while True:
                try:
                    response = chat.send_message(
                        content=google_messages[-1]['parts'], generation_config=self.generation_config(return_json, response_format)
                    )
                    break
                except google.api_core.exceptions.ResourceExhausted:
                    print('Gemini exhausted, retrying...')
                    time.sleep(1)

        input_tokens = response.usage_metadata.prompt_token_count
        output_tokens = response.usage_metadata.total_token_count - input_tokens
        if return_json or response_format:
            response = json.loads(response.text)
        else:
            response = response.text
        return response, input_tokens, output_tokens

    def chat_async(self, messages: list[Message]) -> [str, str]:

        google_messages = transform_messages(messages, False) # Was: messages[:-1]
        try:
            # Initialize the streaming response
            response = self.client.generate_content(google_messages, stream=True,
                                                    generation_config=self.generation_config(False))

            # Collect the streamed parts
            full_response = ''
            for chunk in response:
                full_response += chunk.text
                yield chunk.text, None   # 2nd parameter is reasoning_content. Not available yet for Gemini yet

            # Ensure the iteration completes
            response.resolve()

        except google.generativeai.types.generation_types.IncompleteIterationError:
            print("Please let the response complete iteration before accessing the final accumulated attributes.")

    def token_count(self, text: str) -> int:
        return self.client.count_tokens(text).total_tokens

    def generation_config(self, return_json=False, response_format=None) -> genai.types.GenerationConfig:
        return genai.types.GenerationConfig(
            response_schema=response_format,
            # specifies format of the JSON requested if response_mime_type is `application/json`
            response_mime_type='application/json' if return_json or response_format else 'text/plain',
            **self.model_params
        )


def transform_messages(messages: list[Message], return_json: bool) -> list[dict]:
    return [google_message(msg, return_json) for msg in messages]


def google_message(msg: Message, return_json) -> dict:
    return {
        'role': 'model' if msg.role == 'assistant' else 'user',
        'parts': [Message.to_pil_image(img) for img in msg.images] + [msg.content]
    }


@contextmanager
def temporary_verbosity(level):
    # Sla het huidige logniveau op
    original_level = absl.logging.get_verbosity()
    try:
        # Wijzig naar het nieuwe logniveau
        absl.logging.set_verbosity(level)
        yield
    finally:
        # Herstel het oorspronkelijke logniveau
        absl.logging.set_verbosity(original_level)


import unittest
from justai.model.message import Message, is_image_url

class TestMessage(unittest.TestCase):
    def test_message_creation(self):
        message = Message("user", "Test content")
        self.assertEqual(message.role, "user")
        self.assertEqual(message.content, "Test content")

    def test_message_to_dict(self):
        message = Message("user", "Test content")
        message_dict = message.to_dict()
        self.assertEqual(message_dict["role"], "user")
        self.assertEqual(message_dict["content"], "Test content")

    def test_is_image_url(self):
        valid_url = "https://example.com/image.jpg"
        invalid_url = "https://example.com/not_an_image.txt"
        self.assertTrue(is_image_url(valid_url))
        self.assertFalse(is_image_url(invalid_url))

if __name__ == '__main__':
    import os
    os.chdir(os.path.join(os.path.dirname(__file__), '..'))
    unittest.main()


from justai.models.basemodel import BaseModel


class ModelFactory:
    @staticmethod
    def create(model_name: str, **kwargs) -> BaseModel:
        if model_name.startswith("gpt") or model_name.startswith("o1") or model_name.startswith("o3"):
            from justai.models.openai_models import OpenAIModel
            return OpenAIModel(model_name, params=kwargs)
        elif model_name.endswith(".gguf"):
            from justai.models.gguf_models import GuffModel
            return GuffModel(model_name, params=kwargs)
        elif model_name.startswith("claude"):
            from justai.models.anthropic_models import AnthropicModel
            return AnthropicModel(model_name, params=kwargs)
        elif model_name.startswith("gemini"):
            from justai.models.google_models import GoogleModel
            return GoogleModel(model_name, params=kwargs)
        elif model_name.startswith("grok"):
            from justai.models.xai_models import XAIModel
            return XAIModel(model_name, params=kwargs)
        elif model_name.startswith("deepseek"):
            from justai.models.deepseek_models import DeepSeekModel
            return DeepSeekModel(model_name, params=kwargs)
        elif model_name.startswith("sonar"):
            from justai.models.perplexity_models import PerplexityModel
            return PerplexityModel(model_name, params=kwargs)
        else:
            raise ValueError(f"Model {model_name} not supported")



import base64
import io
import json
import re

import httpx
from PIL import Image


class Message:
    """ Handles the completion as returned by GPT """

    def __init__(self, role=None, content=None, images: list=[], tool_use:dict={}):
        self.role = role
        if isinstance(content, str):
            self.content = content
        else:
            try:
                self.content = json.dumps(content)
            except (TypeError, OverflowError, ValueError, RecursionError):
                raise ValueError("Invalid content type in message. Must be str or json serializable data.")
        self.images = images
        self.tool_use = tool_use

    @classmethod
    def from_dict(cls, data: dict):
        message = cls()
        for key, value in data.items():
            setattr(message, key, value)
        return message

    @staticmethod
    def get_image_type(image):
        if isinstance(image, str) and is_image_url(image):
            return 'image_url'
        elif isinstance(image, bytes):
            return 'image_data'
        elif isinstance(image, Image.Image):
            return 'pil_image'
        else:
            raise ValueError("Unknown content type in message. Must be image url or PIL image or image data.")

    def __bool__(self):
        return bool(self.content)

    def __str__(self):
        res = f'role: {self.role}'
        res += f' content: {self.content}'
        if self.images:
            res += f' [{len(self.images)} images]'
        return res

    def to_dict(self):
        dictionary = {}
        for key, value in self.__dict__.items():
            if value is not None:
                dictionary[key] = value
        return dictionary

    @staticmethod
    def to_base64_image(image):
        image_type = Message.get_image_type(image)
        match image_type:
            case 'image_url':
                img = httpx.get(image).content
            case 'image_data':
                img = image
            case 'pil_image':
                buffered = io.BytesIO()
                image.save(buffered, format="jpeg")
                img = buffered.getvalue()
            case _:
                raise ValueError(f"Unknown image type: {image_type}")
        return base64.b64encode(img).decode("utf-8")

    @staticmethod
    def to_pil_image(image):
        image_type = Message.get_image_type(image)
        match image_type:
            case 'image_url':
                return Image.open(io.BytesIO(httpx.get(image).content))
            case 'image_data':
                return Image.open(io.BytesIO(image))
            case 'pil_image':
                return image
            case _:
                raise ValueError(f"Unknown image type: {image_type}")


def is_image_url(url):
    image_extensions = ('.jpg', '.jpeg', '.png', '.gif', '.bmp', '.tiff', '.webp', '.svg')
    url_pattern = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https:// or ftp:// or ftps://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    if re.match(url_pattern, url):
        if url.lower().endswith(image_extensions):
            return True
    return False



""" Implementation of the OpenAI models. 

Feature table:
    - Async chat:       YES
    - Return JSON:      YES
    - Structured types: YES, via Pydantic  TODO: Add support for native Python types
    - Token count:      YES
    - Image support:    YES
    - Tool use:         not yet

Supported parameters:    
    `# The maximum number of tokens to generate in the completion.
    # Defaults to 16
    # The token count of your prompt plus max_tokens cannot exceed the model's context length.
    # Most models have a context length of 2048 tokens (except for the newest models, which support 4096).
    self.model_params['max_tokens'] = params.get('max_tokens', 800)

    # What sampling temperature to use, between 0 and 2.
    # Higher values like 0.8 will make the output more random, while lower values like 0.2
    # will make it more focused and deterministic.
    # We generally recommend altering this or top_p but not both
    # Defaults to 1
    self.model_params['temperature'] = params.get('temperature', 0.5)

    # An alternative to sampling with temperature, called nucleus sampling,
    # where the model considers the results of the tokens with top_p probability mass.
    # So 0.1 means only the tokens comprising the top 10% probability mass are considered.
    # We generally recommend altering this or temperature but not both.
    # Defaults to 1
    self.model_params['top_p'] = params.get('top_p', 1)

    # How many completions to generate for each prompt.
    # Because this parameter generates many completions, it can quickly consume your token quota.
    # Use carefully and ensure that you have reasonable settings for max_tokens.
    self.model_params['n'] = params.get('n', 1)

    # Number between -2.0 and 2.0.
    # Positive values penalize new tokens based on whether they appear in the text so far,
    # increasing the model's likelihood to talk about new topics.
    # Defaults to 0
    self.model_params['presence_penalty'] = params.get('presence_penalty', 0)

    # Number between -2.0 and 2.0.
    # Positive values penalize new tokens based on their existing frequency in the text so far,
    # decreasing the model's likelihood to repeat the same line verbatim.
    # Defaults to 0
    self.model_params['frequency_penalty'] = params.get('frequency_penalty', 0)
"""

import json
import os

import tiktoken
from dotenv import dotenv_values
from openai import OpenAI, NOT_GIVEN, APIConnectionError, \
    RateLimitError, APITimeoutError, AuthenticationError, PermissionDeniedError, BadRequestError

from justai.model.message import Message
from justai.tools.display import color_print, ERROR_COLOR, DEBUG_COLOR1, DEBUG_COLOR2
from justai.models.basemodel import BaseModel, ConnectionException, AuthorizationException, \
    ModelOverloadException, RatelimitException, BadRequestException, GeneralException


class OpenAIModel(BaseModel):
    def __init__(self, model_name: str, params: dict = None):
        system_message = f"You are {model_name}, a large language model trained by OpenAI."
        super().__init__(model_name, params, system_message)

        # Authentication
        api_key = params.get("OPENAI_API_KEY") or os.getenv("OPENAI_API_KEY") or dotenv_values()["OPENAI_API_KEY"]
        if not api_key:
            color_print("No OpenAI API key found. Create one at https://platform.openai.com/account/api-keys and " +
                        "set it in the .env file like OPENAI_API_KEY=here_comes_your_key.", color=ERROR_COLOR)
        self.client = OpenAI(api_key=api_key)

    def chat(self, messages: list[Message], tools: list, return_json: bool, response_format, use_cache: bool = False) -> tuple[str, int, int, dict|None]:
        # OpenAI models like to have  the system message as part of the conversation
        messages = [Message('system', self.system_message)] + messages

        if self.debug:
            color_print("\nRunning completion with these messages", color=DEBUG_COLOR1)
            [color_print(m, color=DEBUG_COLOR1) for m in messages if hasattr(m, 'text')]
            print()

        try:
            completion = self.completion(messages, tools, return_json, response_format)
        except APIConnectionError as e:
            raise ConnectionException(e)
        except APITimeoutError as e:
            raise ModelOverloadException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            raise AuthorizationException(e)
        except RateLimitError as e:
            raise RatelimitException(e)
        except BadRequestError as e:
            raise BadRequestException(e)
        except Exception as e:
            raise GeneralException(e)

        message = completion.choices[0].message
        message_text = message.content

        # Tool use
        # if completion.choices[0].finish_reason == 'tool_calls':
        #     f = completion.choices[0].message.tool_calls[0].function
        #     tool_use = {
        #         "function_to_call": f.name,
        #         "function_parameters": json.loads(f.arguments),
        #         "call_id": completion.id,
        #     }
        # else:
        #     tool_use = {}
        tool_use = {}

        # Token counts
        input_token_count = completion.usage.prompt_tokens
        output_token_count = completion.usage.completion_tokens

        if message_text and message_text.startswith('```json'):
            print('Unexpected JSON response found in OpenAI completion')
            message_text = message_text[7:-3]
        if self.debug:
            color_print(f"{message_text}", color=DEBUG_COLOR2)

        if response_format and completion.choices[0].message.parsed:
            result = completion.choices[0].message.parsed
        else:
            result = json.loads(message_text) if return_json else message_text
        return result, input_token_count, output_token_count, tool_use
    
    def chat_async(self, messages: list[Message]):
        try:
            completion = self.completion(messages, stream=True)
        except APIConnectionError as e:
            raise ConnectionException(e)
        except (AuthenticationError, PermissionDeniedError) as e:
            raise AuthorizationException(e)
        except APITimeoutError as e:
            raise ModelOverloadException(e)
        except RateLimitError as e:
            raise RatelimitException(e)
        except BadRequestError as e:
            raise BadRequestException(e)
        except Exception as e:
            raise GeneralException(e)

        for item in completion:
            content = item.choices[0].delta.content if hasattr(item.choices[0].delta, "content") else None
            reasoning = item.choices[0].delta.reasoning_content if hasattr(item.choices[0].delta, "reasoning_content") else None
            if content or reasoning:
                yield content, reasoning
               
    def completion(self, messages: list[Message], tools, return_json: bool = False, response_format: BaseModel = None,
                   stream: bool = False):
        transformed_messages = self.transform_messages(messages)
        
        if response_format:
            if stream:
                raise NotImplementedError("streaming is not supported with response_format")
            if tools:
                raise NotImplementedError("tools is not supported with response_format")
            return self.client.beta.chat.completions.parse(
                model=self.model_name,
                messages=transformed_messages,
                response_format=response_format,
                **self.model_params
            )
        else:
            return self.client.chat.completions.create(
                model=self.model_name,
                messages=transformed_messages,
                #tools=tools,
                response_format={"type": "json_object"} if return_json else NOT_GIVEN,
                stream=stream,
                **self.model_params
            )
    
    @staticmethod
    def transform_messages(messages: list[Message]) -> list[dict]:
        def create_openai_message(message):
            msg = {"role": message.role}
            if message.tool_use:
                msg['name'] = message.tool_use['function_to_call']
                msg['content'] = message.tool_use['function_result']
            else:
                content = [{"type": "text", "text": message.content}]
                for image in message.images:
                    content += [{
                                    "type": "image_url",
                                    "image_url": {'url': f"data:image/jpeg;base64,{Message.to_base64_image(image)}"}
                               }]
                msg["content"] = content
            return msg

        result = [create_openai_message(msg) for msg in messages]
        return result

    @staticmethod
    def tool_use_message(tool_use) -> Message:
        return Message(role='function', content='', tool_use=tool_use)

    def token_count(self, text: str) -> int:
        encoding = tiktoken.encoding_for_model(self.model_name)
        return len(encoding.encode(text))



from justai.model.model import Model
from justai.tools.prompts import get_prompt, set_prompt_file, add_prompt_file

if __name__ == '__main__':
    # Ondertaande om de voorkomen dat import optimizer ze leeg gooit
    a = Model
    g = get_prompt
    s = set_prompt_file
    apf = add_prompt_file

